from flask import Blueprint
from flask_restful import Api, Resource

from metrics_layer.api.auth_utils import authenticate_restful
from metrics_layer.core import MetricsLayerConnection
from metrics_layer.core.sql.query_errors import ParseError

metrics_blueprint = Blueprint("metrics", __name__, url_prefix="/api/v1")
api = Api(metrics_blueprint)


class Metrics(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self, sub, metric_name: str):
        try:
            conn = MetricsLayerConnection()
            metric = conn.get_metric(metric_name)
        except ParseError:
            return {"status": "failure", "message": f"metric {metric_name} not found"}, 404

        return {"status": "success", "data": metric.to_dict()}, 200


class MetricsList(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self, sub):
        conn = MetricsLayerConnection()
        metrics = conn.list_metrics()
        metrics_json = [m.to_dict() for m in metrics]
        return {"status": "success", "data": metrics_json}, 200


class Dimensions(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self, sub, dimension_name: str):
        try:
            conn = MetricsLayerConnection()
            dimension = conn.get_dimension(dimension_name)
        except ParseError:
            return {"status": "failure", "message": f"dimension {dimension_name} not found"}, 404

        return {"status": "success", "data": dimension.to_dict()}, 200


class DimensionsList(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self, sub):
        conn = MetricsLayerConnection()
        dimensions = conn.list_dimensions()
        dimensions_json = [d.to_dict() for d in dimensions]
        return {"status": "success", "data": dimensions_json}, 200


api.add_resource(MetricsList, "/metrics")
api.add_resource(Metrics, "/metrics/<metric_name>")
api.add_resource(DimensionsList, "/dimensions")
api.add_resource(Dimensions, "/dimensions/<dimension_name>")



import datetime

import jwt
from flask import current_app

from metrics_layer.api import bcrypt, db
from metrics_layer.api.models.base_model import CoreMixin


class User(db.Model, CoreMixin):
    id = db.Column(db.Integer, unique=True, primary_key=True)
    email = db.Column(db.String(255), unique=True, nullable=False)
    first_name = db.Column(db.String(255))
    last_name = db.Column(db.String(255))
    password = db.Column(db.String(255))
    active = db.Column(db.Boolean())

    def __repr__(self):
        return f"<User email={self.email}>"

    def __init__(self, email, password, first_name=None, last_name=None, active=True):
        self.email = email
        self.password = bcrypt.generate_password_hash(password).decode()
        self.first_name = first_name
        self.last_name = last_name
        self.active = active

    def encode_auth_token(self, user_id):
        """Generates the auth token"""
        try:
            payload = {
                "exp": datetime.datetime.utcnow()
                + datetime.timedelta(
                    days=current_app.config.get("TOKEN_EXPIRATION_DAYS"),
                    seconds=current_app.config.get("TOKEN_EXPIRATION_SECONDS"),
                ),
                "iat": datetime.datetime.utcnow(),
                "sub": user_id,
            }
            return jwt.encode(payload, current_app.config.get("SECRET_KEY"), algorithm="HS256")
        except Exception as e:
            return e

    @staticmethod
    def decode_auth_token(auth_token):
        """
        Decodes the auth token - :param auth_token: - :return: integer|string
        """
        try:
            payload = jwt.decode(auth_token, current_app.config.get("SECRET_KEY"), algorithms=["HS256"])
            return payload["sub"]
        except jwt.ExpiredSignatureError:
            return "Signature expired. Please log in again."
        except jwt.InvalidTokenError:
            return "Invalid token. Please log in again."

    def to_json(self):
        return {
            "id": self.id,
            "email": self.email,
            "active": self.active,
            "first_name": self.first_name,
            "last_name": self.last_name,
        }



from flask import Blueprint, request
from flask_restful import Api, Resource

from metrics_layer.api.auth_utils import authenticate_restful
from metrics_layer.core import MetricsLayerConnection

sql_blueprint = Blueprint("sql", __name__, url_prefix="/api/v1")
api = Api(sql_blueprint)


class ConvertApi(Resource):
    method_decorators = {"post": [authenticate_restful]}

    def post(self, sub):
        request_json = request.get_json()

        if "query" not in request_json:
            return {"status": "failure", "message": "Request missing required parameter 'query'"}, 400

        conn = MetricsLayerConnection()
        converted_query = conn.get_sql_query(sql=request_json["query"])
        return {"status": "success", "data": converted_query}


class QueryApi(Resource):
    method_decorators = {"post": [authenticate_restful]}

    def post(self, sub):
        request_json = request.get_json()

        conn = MetricsLayerConnection()
        df = conn.query(
            sql=request_json.get("sql"),
            metrics=request_json.get("metrics", []),
            dimensions=request_json.get("dimensions", []),
            where=request_json.get("where", []),
            having=request_json.get("having", []),
            order_by=request_json.get("order_by", []),
        )
        return {"status": "success", "data": df.to_dict("records")}


api.add_resource(ConvertApi, "/convert")
api.add_resource(QueryApi, "/query")



import os

BASE_PATH = os.path.dirname(__file__)


class BaseConfig:
    """Base configuration"""

    TESTING = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SECRET_KEY = os.getenv("SECRET_KEY")
    SQLALCHEMY_DATABASE_URI = os.getenv("DATABASE_URL")
    BCRYPT_LOG_ROUNDS = 13
    TOKEN_EXPIRATION_DAYS = 30
    TOKEN_EXPIRATION_SECONDS = 0


class TestingConfig(BaseConfig):
    """Testing configuration"""

    TESTING = True
    PRESERVE_CONTEXT_ON_EXCEPTION = False
    SECRET_KEY = "test-key"
    _default_testing_uri = "sqlite:///" + os.path.join(BASE_PATH, "test.db")
    SQLALCHEMY_DATABASE_URI = os.environ.get("DATABASE_TEST_URL", _default_testing_uri)
    BCRYPT_LOG_ROUNDS = 4
    TOKEN_EXPIRATION_DAYS = 0
    TOKEN_EXPIRATION_SECONDS = 3



from .health_bp import health_blueprint  # noqa
from .metrics_bp import metrics_blueprint  # noqa
from .root_bp import root_blueprint  # noqa
from .sql_bp import sql_blueprint  # noqa
from .users_bp import users_blueprint  # noqa



from functools import wraps

from flask import request

from metrics_layer.api.models import User


def authenticate_restful(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        failure_response = {"status": "failure", "message": "Provide a valid auth token."}

        auth_header = request.headers.get("Authorization")
        if not auth_header:
            return failure_response, 403

        auth_token = auth_header.split(" ")[1]
        sub = User.decode_auth_token(auth_token)

        if isinstance(sub, str):
            failure_response["message"] = sub
            return failure_response, 401

        user = User.query.filter_by(id=sub).first()

        if not user or not user.active:
            return failure_response, 401
        return f(sub, *args, **kwargs)

    return decorated_function



from flask import Blueprint
from flask_restful import Api, Resource

root_blueprint = Blueprint("root", __name__)
api = Api(root_blueprint)


class Root(Resource):
    def get(self):
        return {"status": "success", "message": "TODO react app"}


api.add_resource(Root, "/")



import logging
import sys

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"
FORMAT = "[%(asctime)s] [%(process)d|%(threadName)10s|%(name)s] [%(levelname)s] %(message)s"


def parse_log_level(log_level):
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])


def setup_logging(log_level=DEFAULT_LEVEL):
    # Turn off any existing handlers
    root = logging.getLogger()
    for h in root.handlers[:]:
        root.removeHandler(h)
        h.close()

    logging.basicConfig(stream=sys.stderr, format=FORMAT, level=parse_log_level(log_level))



from sqlalchemy.sql import func

from metrics_layer.api import db


class CoreMixin(object):
    last_modified = db.Column(
        db.DateTime, server_default=func.now(), onupdate=func.current_timestamp(), nullable=False
    )
    created_date = db.Column(db.DateTime, server_default=func.now(), nullable=False)

    @classmethod
    def create(cls, **kwargs):
        obj = cls(**kwargs)
        db.session.add(obj)
        db.session.commit()
        return obj

    @classmethod
    def modify(cls, id, update_dict):
        obj = cls.query.get(id)
        if obj is None:
            return
        else:
            for attr, value in update_dict.items():
                setattr(obj, attr, value)
            db.session.commit()
            return obj

    @classmethod
    def delete(cls, id):
        obj = cls.query.get(id)
        if obj is None:
            return
        else:
            db.session.delete(obj)
            db.session.commit()
            return obj



from flask import Blueprint
from flask_restful import Api, Resource

health_blueprint = Blueprint("health", __name__, url_prefix="/api/v1")
api = Api(health_blueprint)


class HealthPing(Resource):
    def get(self):
        return {"status": "success", "message": "MetricsLayer server active!"}


api.add_resource(HealthPing, "/health")



import logging

from flask import Flask, jsonify
from flask_bcrypt import Bcrypt
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

from metrics_layer.api.logging_utils import FORMAT, setup_logging

setup_logging()

logger = logging.getLogger("metrics_layer.api")

db = SQLAlchemy()
migrate = Migrate()
bcrypt = Bcrypt()


def create_app(config="metrics_layer.api.api_config.BaseConfig"):

    # Instantiate the app
    app = Flask(__name__)

    # Setup logging
    log_handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt=FORMAT)
    log_handler.setFormatter(formatter)
    logger.addHandler(log_handler)

    # Set config
    app.config.from_object(config)

    # Set extensions
    db.init_app(app)
    migrate.init_app(app, db)
    bcrypt.init_app(app)

    # Register blueprints
    import metrics_layer.api.blueprints as bp

    app.register_blueprint(bp.health_blueprint)
    app.register_blueprint(bp.sql_blueprint)
    app.register_blueprint(bp.users_blueprint)
    app.register_blueprint(bp.metrics_blueprint)
    app.register_blueprint(bp.root_blueprint)

    @app.errorhandler(500)
    def internal_error(exception):
        logger.info(f"Error: {exception}")
        return jsonify({"error": True, "code": str(exception)}), 500

    # shell context for flask cli
    @app.shell_context_processor
    def ctx():
        return {"app": app, "db": db}

    return app



import re

from .base import MetricsLayerBase
from .field import Field
from .set import Set


class View(MetricsLayerBase):
    def __init__(self, definition: dict = {}, project=None) -> None:
        if "sql_table_name" in definition:
            definition["sql_table_name"] = self.resolve_sql_table_name(
                definition["sql_table_name"], project.looker_env
            )

        if "sets" not in definition:
            definition["sets"] = []
        self.project = project
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "fields"]
        for k in required_keys:
            if k not in definition:
                raise ValueError(f"View missing required key {k}")

    def printable_attributes(self):
        to_print = ["name", "type", "label", "group_label", "sql_table_name", "number_of_fields"]
        attributes = self.to_dict()
        attributes["number_of_fields"] = f'{len(attributes.get("fields", []))}'
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    @property
    def primary_key(self):
        return next((f for f in self.fields() if f.primary_key == "yes"), None)

    def collect_errors(self):
        fields = self.fields(show_hidden=True)
        field_errors = []
        for field in fields:
            field_errors.extend(field.collect_errors())

        if self.primary_key is None:
            primary_key_error = (
                f"Warning: The view {self.name} does not have a primary key, "
                "specify one using the tag primary_key: yes"
            )
            field_errors += [primary_key_error]

        return field_errors

    def referenced_fields(self):
        fields = self.fields(show_hidden=True)
        result = []
        for field in fields:
            all_fields = [field] + field.get_referenced_sql_query(strings_only=False)
            result.extend(all_fields)
        return result

    def fields(self, show_hidden: bool = True, expand_dimension_groups: bool = False) -> list:
        all_fields = self._valid_fields(expand_dimension_groups=expand_dimension_groups)
        if show_hidden:
            return all_fields
        return [field for field in all_fields if field.hidden == "no" or not field.hidden]

    def _valid_fields(self, expand_dimension_groups: bool):
        if expand_dimension_groups:
            fields = []
            for f in self._definition.get("fields", []):
                field = Field(f, view=self)
                if field.field_type == "dimension_group" and field.timeframes:
                    for timeframe in field.timeframes:
                        if timeframe == "raw":
                            continue
                        field.dimension_group = timeframe
                        fields.append(field)
                elif field.field_type == "dimension_group" and field.intervals:
                    for interval in field.intervals:
                        field.dimension_group = f"{interval}s"
                        fields.append(field)
                else:
                    fields.append(field)
        else:
            fields = [Field(f, view=self) for f in self._definition.get("fields", [])]

        return fields

    def _field_name_to_remove(self, field_expr: str):
        # Skip the initial - sign
        field_clean_expr = field_expr[1:]
        if "." in field_clean_expr:
            view_name, field_name = field_clean_expr.split(".")
            if view_name == self.name:
                return field_name
            return None
        return field_clean_expr

    def resolve_sql_table_name(self, sql_table_name: str, looker_env: str):
        start_cond, end_cond = "-- if", "--"
        if start_cond in sql_table_name:
            # Find the condition that is chosen in the looker env
            conditions = re.findall(f"{start_cond}([^{end_cond}]*){end_cond}", sql_table_name)
            try:
                condition = next((cond for cond in conditions if cond.strip() == looker_env))
            except StopIteration:
                raise ValueError(
                    f"""Your sql_table_name: '{sql_table_name}' contains a conditional and
                    we could not match that to the conditional value you passed: {looker_env}"""
                )

            full_phrase = start_cond + condition + end_cond

            # Use regex to extract the value associated with the condition
            searchable_sql_table_name = sql_table_name.replace("\n", "")
            everything_between = f"{full_phrase}([^{end_cond}]*){end_cond}"
            everything_after = f"(?<={full_phrase}).*"
            result = re.search(everything_between, searchable_sql_table_name)
            if result:
                return result.group().replace(end_cond, "").strip()

            result = re.search(everything_after, searchable_sql_table_name)
            return result.group().strip()

        return sql_table_name

    def list_sets(self):
        return [Set({**s, "view_name": self.name}, project=self.project) for s in self.sets]

    def get_set(self, set_name: str):
        return next((s for s in self.list_sets() if s.name == set_name), None)



from flask import Blueprint, request
from flask_restful import Api, Resource
from sqlalchemy import exc

from metrics_layer.api import bcrypt, db
from metrics_layer.api.auth_utils import authenticate_restful
from metrics_layer.api.models import User

users_blueprint = Blueprint("users", __name__, url_prefix="/api/v1")
api = Api(users_blueprint)


class Users(Resource):
    method_decorators = {
        "get": [authenticate_restful],
        "patch": [authenticate_restful],
        "delete": [authenticate_restful],
    }

    def get(self, user_id: str):
        """Get single user details"""
        response = {"status": "failure", "message": "User does not exist"}
        try:
            user = User.query.filter_by(id=int(user_id)).first()
            if not user:
                return response, 404
            return {"status": "success", "data": user.to_json()}, 200

        except ValueError:
            return response, 404

    def patch(self, user_id: str):
        request_json = request.get_json()
        user = User.modify(user_id, request_json)
        return {"status": "success", "data": user.to_json()}, 200

    def delete(self, user_id: str):
        user = User.delete(user_id)
        if not user:
            return {"status": "failure", "message": "User does not exist"}, 404
        return {"status": "success"}


class UsersList(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self):
        """Get all users"""
        response_object = {
            "status": "success",
            "data": {"users": [user.to_json() for user in User.query.all()]},
        }
        return response_object, 200


class Register(Resource):
    def post(self):
        post_data = request.get_json()
        failure_response = {"status": "failure", "message": "Invalid payload."}
        if not post_data:
            return failure_response, 400

        email = post_data.get("email")
        password = post_data.get("password")  # new

        try:
            user = User.query.filter_by(email=email).first()
            if not user:
                new_user = User.create(email=email, password=password)
                auth_token = new_user.encode_auth_token(new_user.id)
                return {"status": "success", "message": f"{email} was added", "auth_token": auth_token}, 200
            else:
                failure_response["message"] = "Sorry. That email already exists."
                return failure_response, 400

        except (exc.IntegrityError, ValueError):
            db.session.rollback()
            return failure_response, 400


class Login(Resource):
    def post(self):
        post_data = request.get_json()
        failure_response = {"status": "failure", "message": "Invalid payload."}
        if not post_data:
            return failure_response, 400

        email = post_data.get("email")
        password = post_data.get("password")

        try:
            # fetch the user data
            user = User.query.filter_by(email=email).first()

            if user and bcrypt.check_password_hash(user.password, password):
                auth_token = user.encode_auth_token(user.id)
                if auth_token:
                    response = {
                        "status": "success",
                        "message": "Successfully logged in.",
                        "auth_token": auth_token,
                    }
                    return response, 200

            else:
                failure_response["message"] = "User does not exist."
                return failure_response, 404

        except Exception:
            failure_response["message"] = "Try again."
            return failure_response, 500


class Logout(Resource):
    method_decorators = {"get": [authenticate_restful]}

    def get(self, sub):
        response = {"status": "success", "message": "Successfully logged out."}
        return response, 200


api.add_resource(Register, "/register")
api.add_resource(Login, "/login")
api.add_resource(Logout, "/logout")
api.add_resource(UsersList, "/users")
api.add_resource(Users, "/users/<user_id>")



from flask.cli import FlaskGroup

from metrics_layer.api import create_app, db
from metrics_layer.api.models import User

cli = FlaskGroup(create_app=create_app)


@cli.command("recreate_db")
def recreate_db():
    db.drop_all()
    db.create_all()
    db.session.commit()


@cli.command("seed_db")
def seed_db():
    """Seeds the database."""
    User.create(
        first_name="First",
        last_name="Last",
        email="test@example.com",
        password="password",
    )


if __name__ == "__main__":
    cli()



import json
import os
from copy import deepcopy

from metrics_layer.core.model.definitions import Definitions


class MetricsLayerConnectionError(Exception):
    pass


class ConnectionType:
    snowflake = Definitions.snowflake
    bigquery = Definitions.bigquery


class BaseConnection:
    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class SnowflakeConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        account: str,
        username: str,
        password: str,
        role: str = None,
        warehouse: str = None,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.snowflake
        self.name = name
        self.account = account
        self.username = username
        self.password = password
        self.role = role
        self.warehouse = warehouse
        self.database = database
        self.schema = schema

    def to_dict(self):
        """Dict for use with the snowflake connector"""
        base = {"user": self.username, "password": self.password, "account": self.account}
        if self.warehouse:
            base["warehouse"] = self.warehouse
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        if self.role:
            base["role"] = self.role
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "account", "user", "database", "schema", "warehouse", "role"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class BigQueryConnection(BaseConnection):
    def __init__(self, name: str, credentials: str, **kwargs) -> None:
        self.type = ConnectionType.bigquery
        self.name = name
        self.credentials = self._convert_json_if_needed(credentials, kwargs)
        self.project_id = self.credentials["project_id"]

    def to_dict(self):
        """Dict for use with the BigQuery connector"""
        return {"credentials": self.credentials, "project_id": self.project_id}

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("credentials")
        attributes["name"] = self.name
        sort_order = ["name", "project_id"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}

    @staticmethod
    def _convert_json_if_needed(creds: dict, kwargs: dict):
        if isinstance(creds, dict):
            return deepcopy(creds)
        elif isinstance(creds, str):
            try:
                return json.loads(creds)
            except json.JSONDecodeError:
                # This means it's a file path not a JSON string
                if os.path.isabs(creds):
                    path = creds
                else:
                    path = os.path.join(kwargs["directory"], creds)

                with open(path, "r") as f:
                    return json.load(f)
        else:
            raise TypeError(f"BigQuery credentials json had wrong type: {type(creds)} for value {creds}")
